/**
 * File:   math3d.h
 * Author: Benjamin Pickhardt (hackCraft.de)
 *
 * Created on Dez 23, 1999, 3:33 PM
 */

#ifndef _MATH3D_H
#define _MATH3D_H

#include <math.h>

//--[Include Files]----------------->
/*
#include <math.h>
#include <stdio.h>
#include <string.h>
*/

#define PI_OVER_180 0.017453292

//--[Functional Macros]----------------->


#define set3i(v, func) loop3i(v[i] = func;)

#define vectorset(v, x, y, z) { (v)[0] = (x); (v)[1] = (y); (v)[2] = (z); }


//--[Machine Specific]-------------->

/*
typedef  float           FLOAT;

typedef  long            INTEG;

typedef  unsigned char   BOOLEAN;

typedef  unsigned long   FLAGS;

typedef  unsigned char   BYTE;

typedef  unsigned short  WORD;

typedef  unsigned long   DWORD;
*/

//--[Non Machine Specific]---------->

/*
typedef  FLOAT           ANGLE;

typedef  FLOAT           SCALAR;

typedef  SCALAR          VECTOR[4]; // homogenous/plane/rgba

typedef  SCALAR          MATRIX[4*4];
*/

//--[Array Accessors]--------------->

#define  X 0
#define  Y 1
#define  Z 2
#define  W 3
#define  D 3

// better try not to use that letters for variables etc.
// ...but who would use that short-uppercase-names anyway?


//--[Misc Defines]------------------>


#define ppercent(zero_pc, hundred_pc, p_pc)  (zero_pc + ((hundred_pc)-(zero_pc))*(p_pc)/100 )

#define inside(min, v, max)    ( ( (min) <= (v) ) && ( (v) <= (max) ) )

#define setflag(a, b)        (a) = ( FLAGS((a)) | FLAGS((b)) )

#define resetflag(a, b)      (a) = ( FLAGS((a)) & ( ~ FLAGS((b)) ) )


//--[Numeric Defines]------------------>

#define xsqr(x)              ( (x) * (x) )

#define xsgn(v)              (((v)<0)?(-1):((v>0)?(+1):(0)))

#define xabs(v)              (((v)<0)?(-v):(v))

#define xmod(a, b)           ((a) -= long((b)) * (long( (a) ) / long( (b) )))

#define xmax(v1,v2)          (((v1)>(v2))?(v1):(v2))

#define xmin(v1,v2)          (((v1)<(v2))?(v1):(v2))

#define xminmax(v, min, max)   ( xmin(xmax(v, max), min) )

//--[Parametric Spline Defines]--->


// -1.0 <= u <= 2.0

#define spline_b1(u)    ((u) * ((u) - 1) * ((u) - 2) / (-6.0F))

#define spline_b2(u)    (((u) + 1) * ((u) - 1) * ((u) - 2) / (2.0F))

#define spline_b3(u)    (((u) + 1) * (u) * ((u) - 2) / (-2.0F))

#define spline_b4(u)    (((u) + 1) * (u) * ((u) - 1) / (6.0F))

#define spline_func(a, b, c, d, u) ((a) * spline_b1((u)) + (b) * spline_b2((u)) + (c) * spline_b3((u)) + (d) * spline_b4((u)))


//--[Geometrical Formula Defines]--->


#define negVector(v)             { v[X] = -v[X]; v[Y] = -v[Y]; v[Z] = -v[Z]; }

#define setVector(v, x,  y, z)   { v[X] = x; v[Y] = y; v[Z] = z; }


//--[return value]-->


#define lineDistanceSqr(A, B)  ( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y])+((A)[Z]-(B)[Z])*((A)[Z]-(B)[Z]) )

#define lineDistance(A, B)   (sqrt( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y])+((A)[Z]-(B)[Z])*((A)[Z]-(B)[Z]) ))


#define manhattanDistance(A, B)       ( xabs((A)[X]-(B)[X]))+xabs((A)[Y]-(B)[Y]))+xabs((A)[Z]-(B)[Z]))  )


// --Following -------------------->
// n = normal of plane;
// p = point of line/ray;
// v = slope of line/ray
// d = "D" from the plane equation


#define calcT(n, p, v, d)            (-(n[X]*p[X]+n[Y]*p[Y]+n[Z]*p[Z]+d)/(n[X]*v[X]+n[Y]*v[Y]+n[Z]*v[Z]))

#define calcP(nrml, pnt, d)          (nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]+d)

#define calcA(nrml, pnt, d)          (-(nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]+d)/pnt[X])

#define calcB(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Z]*pnt[Z]+d)/pnt[Y])

#define calcC(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+d)/pnt[Z])

#define calcD(nrml, pnt)             (-(nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]))

#define calcX(nrml, pnt, d)          (-(nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]+d)/nrml[X])

#define calcY(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Z]*pnt[Z]+d)/nrml[Y])

#define calcZ(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+d)/nrml[Z])

#define magnitude(A)                 ( sqrt( xsqr((A)[X]) + xsqr((A)[Y]) + xsqr((A)[Z]) ) )

#define dotprod(A, B)                ((A)[X] * (B)[X] + (A)[Y] * (B)[Y] + (A)[Z] * (B)[Z] )


#define clip(a1, b1, a2, b2, clp)    ( (a1) - ( (a1) - (a2) ) / ( (b1) - (b2) ) * ( (b1) - (clp) ) )

#define clif(b1, b2, f)              ( (b1) - ( (b1) - (b2) ) * (f) )

#define clic(b1, b2, f, d)           ( (b1) - ( (b1) - (b2) ) * (f) / (d) )


//--[no return value]-->

#define cross(result, A, B)         { (result)[X] = (A)[Y] * (B)[Z] - (A)[Z] * (B)[Y]; (result)[Y] = (A)[Z] * (B)[X] - (A)[X] * (B)[Z]; (result)[Z] = (A)[X] * (B)[Y] - (A)[Y] * (B)[X]; }

#define vectoradd(result, A, B)     { (result)[X] = (A)[X] + (B)[X]; (result)[Y] = (A)[Y] + (B)[Y]; (result)[Z] = (A)[Z] + (B)[Z]; }

#define vectorsub(result, A, B)     { (result)[X] = (A)[X] - (B)[X]; (result)[Y] = (A)[Y] - (B)[Y]; (result)[Z] = (A)[Z] - (B)[Z]; }

#define scalarmul(result, A, B)     { (result)[X] = (A)[X] * (B); (result)[Y] = (A)[Y] * (B); (result)[Z] = (A)[Z] * (B); }

#define scalardiv(result, A, B)     { (result)[X] = (A)[X] / (B); (result)[Y] = (A)[Y] / (B); (result)[Z] = (A)[Z] / (B); }

#define normalize(result, A)        { scalardiv( (result), (A), magnitude((A)) ); }

// Prints a 3f vector.
#define printVector(vec) \
{ \
    printf("< %3.2f %3.2f %3.2f >\n", (vec)[0], (vec)[1], (vec)[2]); \
}


//--[misc vec/mtx functions]-->


// Intersection of a Line (p0,p1) with the Plane (normal,d) result point exists if result is non-zero.
// sectionLinePlane(int result, VECTOR normal, float d, VECTOR p0, VECTOR p1, VECTOR resultp)
#define sectionLinePlane(result, normal, d, p0, p1, resultp) \
{ \
  (result) = 0 \
  float d0 = calcP((normal), (p0), (d)); \
  float d1 = calcP((normal), (p1), (d)); \
  if ( xsgn(d0) != xsgn(d1) ) { \
    float dd = d0 / (d0 - d1); \
    (resultp)[X] = (p0)[X] - (((p0)[X] - (p1)[X]) * dd); \
    (resultp)[Y] = (p0)[Y] - (((p0)[Y] - (p1)[Y]) * dd); \
    (resultp)[Z] = (p0)[Z] - (((p0)[Z] - (p1)[Z]) * dd); \
    result = 1; \
  } \
}


// Make normal from three Points (VECTOR normal, VECTOR p0, VECTOR p1, VECTOR p2)
#define makeNormal(result, p0, p1, p2) \
{ \
    float v0[3]; \
    float v1[3]; \
    vectorsub(v0, (p0), (p1)); \
    vectorsub(v1, (p1), (p2)); \
    cross((result), v0, v1); \
}


// Prints an Matrix of gl-Format (as used here)
#define printMatrix(glMatrix) \
{ \
    printf("/ %3.2f %3.2f %3.2f %3.2f \\\n", (glMatrix)[0], (glMatrix)[4], (glMatrix)[ 8], (glMatrix)[12]); \
    printf("| %3.2f %3.2f %3.2f %3.2f |\n", (glMatrix)[1], (glMatrix)[5], (glMatrix)[ 9], (glMatrix)[13]); \
    printf("| %3.2f %3.2f %3.2f %3.2f |\n", (glMatrix)[2], (glMatrix)[6], (glMatrix)[10], (glMatrix)[14]); \
    printf("\\ %3.2f %3.2f %3.2f %3.2f /\n", (glMatrix)[3], (glMatrix)[7], (glMatrix)[11], (glMatrix)[15]); \
}


// Transposes a 4x4 Matrix - useful to invert ortho-normal-bases.
#define transposeMatrix4(m) { float tmp; loopi(4) { loopj(i) { tmp = m[4*i+j]; m[4*i+j] = m[4*j+i]; m[4*j+i] = tmp; } } }


// Set Matrix Values (MATRIX result, float a1, float ...)
#define setMatrix(result,  a1, a2, a3, a4,  b1, b2, b3, b4,  c1, c2, c3, c4,  d1, d2, d3, d4) \
{ \
  (result)[ 0]=(a1); (result)[ 1]=(a2); (result)[ 2]=(a3); (result)[ 3]=(a4); \
  (result)[ 4]=(b1); (result)[ 5]=(b2); (result)[ 6]=(b3); (result)[ 7]=(b4); \
  (result)[ 8]=(c1); (result)[ 9]=(c2); (result)[10]=(c3); (result)[11]=(c4); \
  (result)[12]=(d1); (result)[13]=(d2); (result)[14]=(d3); (result)[15]=(d4); \
}


// Set Identity Matrix (MATRIX result)
#define setIdentMatrix(result) \
{ \
  setMatrix((result), 1,0,0,0, \
                      0,1,0,0, \
                      0,0,1,0, \
                      0,0,0,1); \
}


// Apply Matrix to Vector
// applyMatrix2(MATRIX M, VECTOR V)
#define applyMatrix2(M, V) \
{ \
    float pre[] = { (V)[0], (V)[1], (V)[2] }; \
    (V)[0] = (M)[0]*(pre)[0] + (M)[4]*(pre)[1] + (M)[ 8]*(pre)[2] + (M)[12]; \
    (V)[1] = (M)[1]*(pre)[0] + (M)[5]*(pre)[1] + (M)[ 9]*(pre)[2] + (M)[13]; \
    (V)[2] = (M)[2]*(pre)[0] + (M)[6]*(pre)[1] + (M)[10]*(pre)[2] + (M)[14]; \
}


// Apply Matrix to Vector and stores to result
// applyMatrix3(VECTOR result, MATRIX M, VECTOR V)
#define applyMatrix3(result, M, V) \
{ \
  (result)[X] = ((V)[0]*(M)[0*4+0]) + ((V)[1]*(M)[1*4+0]) + ((V)[2]*(M)[2*4+0]) + ((M)[3*4+0]); \
  (result)[Y] = ((V)[0]*(M)[0*4+1]) + ((V)[1]*(M)[1*4+1]) + ((V)[2]*(M)[2*4+1]) + ((M)[3*4+1]); \
  (result)[Z] = ((V)[0]*(M)[0*4+2]) + ((V)[1]*(M)[1*4+2]) + ((V)[2]*(M)[2*4+2]) + ((M)[3*4+2]); \
}


// Multiplies Matrices M and N and stores result to M
// mulMatrix(MATRIX M, MATRIX N)
#define mulMatrix(M, N) \
{ \
  MATRIX tmpM; \
  for (int j=0; j<4; j++) { \
    for (int i=0; i<4; i++) { \
      tmpM[i+4*j] = ((N)[i+4*0]*(M)[0+4*j]) + ((N)[i+4*1]*(M)[1+4*j])+ \
                    ((N)[i+4*2]*(M)[2+4*j]) + ((N)[i+4*3]*(M)[3+4*j]); \
    }; \
  }; \
  memcpy((M), tmpM, sizeof(MATRIX)); \
}


// Add Matrices M and N and stores result to M
// addMatrix(MATRIX M, MATRIX N)
#define addMatrix(M, N) \
{ \
    for (int i=0; i<16; i++) \
      (M)[i] += (N)[i]; \
}


// Rotate Matrix M around X-Axis
// RXMatrix(MATRIX M, ANGLE x)
#define RXMatrix(M, x) \
{ \
  MATRIX N; \
  float c = cos((x)*PI_OVER_180); \
  float s = sin((x)*PI_OVER_180); \
  setMatrix(N, 1, 0, 0, 0, \
               0, c, s, 0, \
               0,-s, c, 0, \
               0, 0, 0, 1); \
  mulMatrix((M), N); \
}


// Rotate Matrix M around Y-Axis
// RYMatrix(MATRIX M, ANGLE y)
#define RYMatrix(M, y) \
{ \
  MATRIX N; \
  float c = cos((y)*PI_OVER_180); \
  float s = sin((y)*PI_OVER_180); \
  setMatrix(N, c, 0,-s, 0, \
               0, 1, 0, 0, \
               s, 0, c, 0, \
               0, 0, 0, 1); \
  mulMatrix((M), N); \
}


// Rotate Matrix M around Z-Axis
// RZMatrix(MATRIX M, ANGLE z)
#define RZMatrix(M, z) \
{ \
  MATRIX N; \
  float c = cos((z)*PI_OVER_180); \
  float s = sin((z)*PI_OVER_180); \
  setMatrix(N, c, s, 0, 0, \
              -s, c, 0, 0, \
               0, 0, 1, 0, \
               0, 0, 0, 1); \
  mulMatrix((M), N); \
}


// Translates Matrix M by (x,y,z)
// TMatrix(MATRIX M, float x, float y, float z)
#define TMatrix(M, x, y, z) \
{ \
  (M)[12] += x; \
  (M)[13] += y; \
  (M)[14] += z; \
}


// Scales Matrix M with factors (x,y,z)
// SMatrix(MATRIX M, float x, float y, float z)
#define SMatrix(M, x, y, z) \
{ \
  MATRIX N; \
  setMatrix(N, (x),  0,  0,  0, \
                 0,(y),  0,  0, \
                 0,  0,(z),  0, \
                 0,  0,  0,  1); \
  mulMatrix(M, N); \
}


// Rotate Vector Around X-Axis (VECTOR V, ANGLE x)
#define RXVector(V, x) \
{ \
  MATRIX M; \
  float c = cos((x)*PI_OVER_180); \
  float s = sin((x)*PI_OVER_180); \
  setMatrix(M, 1, 0, 0, 0, \
               0, c, s, 0, \
               0,-s, c, 0, \
               0, 0, 0, 1); \
  applyMatrix(M, (V)); \
}


// Rotate Vector Around Y-Axis (VECTOR V, ANGLE y)
#define RYVector(V, y) \
{ \
  MATRIX M; \
  float c = cos((y)*PI_OVER_180); \
  float s = sin((y)*PI_OVER_180); \
  setMatrix(M, c, 0,-s, 0, \
               0, 1, 0, 0, \
               s, 0, c, 0, \
               0, 0, 0, 1); \
  applyMatrix(M, (V)); \
}


// Rotate Vector Around Z-Axis (VECTOR V, ANGLE z)
#define RZVector(V, z) \
{ \
  MATRIX M; \
  float c = cos((z)*PI_OVER_180); \
  float s = sin((z)*PI_OVER_180); \
  setMatrix(M, c, s, 0, 0, \
              -s, c, 0, 0, \
               0, 0, 1, 0, \
               0, 0, 0, 1); \
  applyMatrix(M, (V)); \
}


#define quat_norm(q0) \
( \
    (q0)[0] * (q0)[0] + (q0)[1] * (q0)[1] + (q0)[2] * (q0)[3] + (q0)[3] * (q0)[3] \
)


#define quat_inv_norm(q0) \
( \
    1.0 / quat_norm(q0) \
)


#define quat_inv(q_, q0, inv_norm) \
{ \
    (q_)[0] = (q0)[0] * inv_norm; \
    (q_)[1] = (q0)[1] * inv_norm; \
    (q_)[2] = (q0)[2] * inv_norm; \
    (q_)[3] = (q0)[3] * inv_norm; \
}


#define quat_add(q_, q0, q1) \
{ \
    (q_)[0] = (q0)[0] + (q1)[0]; \
    (q_)[1] = (q0)[1] + (q1)[1]; \
    (q_)[2] = (q0)[2] + (q1)[2]; \
    (q_)[3] = (q0)[3] + (q1)[3]; \
}


#define quat_mul(q_, q0, q1) \
{ \
  (q_)[3] = (q0)[3] * (q1)[3] - (q0)[0] * (q1)[0] - (q0)[1] * (q1)[1] - (q0)[2] * (q1)[2]; \
  (q_)[0] = (q0)[3] * (q1)[0] + (q0)[0] * (q1)[3] + (q0)[1] * (q1)[2] - (q0)[2] * (q1)[1]; \
  (q_)[1] = (q0)[3] * (q1)[1] + (q0)[1] * (q1)[3] + (q0)[2] * (q1)[0] - (q0)[0] * (q1)[2]; \
  (q_)[2] = (q0)[3] * (q1)[2] + (q0)[2] * (q1)[3] + (q0)[0] * (q1)[1] - (q0)[1] * (q1)[0]; \
}


#endif
