/**
 * File:   cBuilding.h
 * Author: Benjamin Pickhardt (hackCraft.de)
 *
 * Created on March 28, 2008, 10:23 PM
 */

#ifndef BUILDING_H
#define BUILDING_H

class cBuilding;

#include "cObject.h"
#include "cWeapon.h"

/**
 * Models a building of varring size.
 */
class cBuilding : public cObject {
protected:
    static int sInstances;
    static long sTextures[1];
protected:
    float mBase[16]; // Base matrix for explosion ie. center.
    cExplosion mExplosion;
    float mRooms[3];
public:
    cBuilding(int x, int y, int z, int rooms_x, int rooms_y, int rooms_z);
    
    virtual rImpact* damageByBeam(float* a, float* b, float radius, float damage, bool all);
    
    virtual void onSpawn();
    
    virtual void animate(float spf);
    virtual void transform();
    virtual void drawSolid();
    virtual void drawEffect();
};


/**
 * Models an urban street-tile of different types.
 */
class cTile : public cObject {
public:
    /// Possible models of tile.
    enum Kinds {
        KIND_CONCRETE,
        KIND_ROAD_EASTWEST,
        KIND_ROAD_NORTHSOUTH,
        KIND_ROAD_JUNKTION,

        MAX_KINDS
    };
protected:
    /// Counts number of instances.
    static int sInstances;
    /// Contains shared texture(s) of all instances.
    static long sTextures[1];
    
    static long sDisplaylists[MAX_KINDS*2];
protected:
    /// This tile has the named appeareance (as in Kinds-enum).
    int mKind;
public:

    /// Initialises a tile of a certain kind at the given tile-grid position.
    cTile(int x, int y, int z, int kind);
    virtual rImpact* damageByBeam(float* a, float* b, float radius, float damage, bool all);
    virtual void drawSolid();
};

#endif
