// cMech.cpp

#include "cObject.h"

#include <vector>
#include <iostream>

#include "cMech.h"
#include "cController.h"
#include "cPad.h"
#include "macros.h"

#include <assert.h>
#include <GL/gl.h>
#define GL_GLEXT_PROTOTYPES
#include <GL/glext.h>
#include <AL/al.h>
#include <AL/alut.h>

DEFINE_Console_printf
DEFINE_glprintf

        int cMech::sInstances = 0;
long cMech::sTextures[1];

cMech::cMech(float* pos, float* rot) {
    sInstances++;
    if (sInstances == 1) {
        // First one initializes
        unsigned int texname;
        const int WIDTH = 64;
        const int HEIGHT = 64;
        const int DEPTH = 64;
        unsigned char texels[WIDTH * HEIGHT * DEPTH * 3];
        //float s = 3;
        //generateTexture3D(WIDTH, HEIGHT, DEPTH, texels, deltaFunc_NoisyPlasma, ColorFunc_Delta,ColorFunc_Delta,ColorFunc_Delta, 1*s,0.2*s,1*s);
        //glBindTexture3D(texname, false,false, true,true,true, WIDTH,HEIGHT,DEPTH,texels);
        long t = 0;
        for (int i = 0; i < WIDTH; i++) {
            for (int j = 0; j < HEIGHT; j++) {
                for (int k = 0; k < DEPTH; k++) {
                    int d = deltaFunc_NoisyPlasma(i, j, k);
                    int r = d, g = d, b = d;
                    if (d > 220) {
                        r += 140;
                        g += 120;
                        b += 100;
                    } else if (d > 200) {
                        r += 120;
                        g += 90;
                        b += 60;
                    }
                    texels[t++] = r;
                    texels[t++] = g;
                    texels[t++] = b;
                }
            }
        }
        glBindTexture3D(texname, true, false, true, true, true, WIDTH, HEIGHT, DEPTH, texels);
        sTextures[0] = texname;
    }

    addRole(COLLIDEABLE);
    addRole(DAMAGEABLE);
    addRole(SOCIALISED);
    addRole(ENTITY);
    addRole(MECH);

    mController = new cController(this, true);

    if (pos != NULL) memcpy(mPos.data(), pos, sizeof (float) * 3);
    if (rot != NULL) memcpy(mRot.data(), rot, sizeof (float) * 3);
    memset(mTwr, 0, sizeof (float) * 3);

    for (int i = 0; i < MAX_ROTATORS; i++) {
        mRotators[i][0] = mRotators[i][1] = mRotators[i][2] = 0.0f;
    };

    mRadius = 1.8f;

    mCurrentWeapon = 0;
    mJetOn = 0;
    mSpeed = 0;
    //mGrounded = true;
    mJumpstate = FALLING;

    mCameraState = 1;

    //
    ALuint buffer;
    //buffer = alutCreateBufferHelloWorld();
    buffer = alutCreateBufferFromFile("data/sound/pow.wav");
    alGenSources(1, &mSoundSource);
    alSourcei(mSoundSource, AL_BUFFER, buffer);
    alSourcef(mSoundSource, AL_PITCH, 1.0f);
    alSourcef(mSoundSource, AL_GAIN, 1.0f);
    alSourcefv(mSoundSource, AL_POSITION, mPos.data());
    alSourcefv(mSoundSource, AL_VELOCITY, mVel.data());
    alSourcei(mSoundSource, AL_LOOPING, AL_FALSE);

    //if (this->mSerial != 20280809030000001ULL) alSourcePlay(mSource);
    alSourcePlay(mSoundSource);

    mConsole = Console_new(50 * 9, 50);
    Console_printf(mConsole, "* Welcome to LinWarrior3D\n");
    mLastMessage = 0;
}

cMech::~cMech() {
    if (sInstances == 1) {
        // Last one cleans up

    }
    sInstances--;
}

void cMech::onSpawn() {
    this->mountWeapon((char*) "CTorsor", &mExplosion, false);
    std::cout << "Mech spawned" << std::endl;
}

void cMech::multEyeMatrix() {
    float camera[16];
    float rot_inv[16];
    float pos_inv[16];

    glPushMatrix();
    {
        // Get Inverse Components of Head Matrix
        glLoadMatrixf(mHead);
        glGetTransposeInverseRotationMatrix(rot_inv);
        glGetInverseTranslationMatrix(pos_inv);
        // Compose Camera Matrix from inverse components
        glLoadIdentity();
        glMultMatrixf(rot_inv);
        glMultMatrixf(pos_inv);
        glGetFloatv(GL_MODELVIEW_MATRIX, camera);
        //printMatrix(camera);

        { // OpenAL Listener setup
            float s = -0.1;
            float vel[] = {
                mVel[0] * cWorld::instance->mSPF*s, mVel[1] * cWorld::instance->mSPF*s, mVel[2] * cWorld::instance->mSPF * s
            };
            float pos[] = {
                mPos[0], mPos[1], mPos[2]
            };
            alListenerfv(AL_POSITION, pos);
            alListenerfv(AL_VELOCITY, vel);
            float at_and_up[] = {
                mHead[8], mHead[9], mHead[10],
                0, -1, 0
            };
            //memcpy(&at_and_up[0], &mHead[8], sizeof(float)*3);
            //memcpy(&at_and_up[3], &mHead[4], sizeof(float)*3);

            alListenerfv(AL_ORIENTATION, at_and_up);
        }
    }
    glPopMatrix();

    if (abs(mCameraState) == 1) { // 1st
        //glRotatef(180, 0, 1, 0);
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 2) { // 3rd
        glTranslatef(0, 0, -5);
        glRotatef(15, 1, 0, 0);
        //glRotatef(180, 0, 1, 0);
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 3) { // 3rd Far
        glTranslatef(0, 0, -15);
        glRotatef(15, 1, 0, 0);
        //glRotatef(180, 0, 1, 0);
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 4) { // Reverse 3rd Far
        glTranslatef(0, 0, -15);
        glRotatef(15, 1, 0, 0);
        //glRotatef(0, 0, 1, 0);
        glRotatef(180, 0, 1, 0);
        glMultMatrixf(camera);
    } else if (abs(mCameraState) == 5) { // Map Near
        glTranslatef(0, 0, -50);
        glRotatef(90, 1, 0, 0);
        //glRotatef(180, 0, 1, 0);
        glMultMatrixf(pos_inv);
    } else if (abs(mCameraState) == 6) { // Map Far
        glTranslatef(0, 0, -100);
        glRotatef(90, 1, 0, 0);
        //glRotatef(180, 0, 1, 0);
        glMultMatrixf(pos_inv);
    }

}

void cMech::Pattern(float t, char *pattern) {
    int l = strlen(pattern);
    int s = int(t / 90.0) % l;
    char c = pattern[s];
    switch (c) {
        case 'l': ChassisLR(+5);
            break;
        case 'r': ChassisLR(-5);
            break;
        case 'n': break;
        default: break;
    };
}

void cMech::ChassisLR(float r) {
    const float limit = 1.75;
    float e = r;
    if (e>+limit) e = +limit;
    if (e<-limit) e = -limit;
    mRot[Y] += e;
    mRot[Y] = fmod(mRot[Y], 360);
    if (mRot[Y] < 0) mRot[Y] += 360;
}

void cMech::ChassisUD(float r) {
    float e = r;
    if (e>+1) e = +1;
    if (e<-1) e = -1;
    mRot[X] += e;
    mRot[X] = fmod(mRot[X], 360);
    if (mRot[X] < 0) mRot[X] += 360;
}

void cMech::TowerLR(float r) {
    float e = r;
    if (e>+2) e = +2;
    if (e<-2) e = -2;
    mTwr[Y] += e;
    if (this->hasRole(cObject::IMMOBILE)) {
        mTwr[Y] = fmod(mTwr[Y], 360);
    } else {
        if (mTwr[Y] > +75) mTwr[Y] = +75;
        if (mTwr[Y] < -75) mTwr[Y] = -75;
    }
}

void cMech::TowerUD(float r) {
    float e = r;
    if (e>+2) e = +2;
    if (e<-2) e = -2;
    mTwr[X] += e;
    if (mTwr[X] > +75) mTwr[X] = +75;
    if (mTwr[X] < -75) mTwr[X] = -75;
}

void cMech::fireAllWeapons() {

    loopi(mWeapons.size()) {
        mWeapons[i]->fire(mTarget);
    }
}

void cMech::fireWeapon(unsigned n) {
    if (n >= mWeapons.size()) return;
    mWeapons[n]->fire(mTarget);
}

void cMech::mountWeapon(char* point, cWeapon *weapon, bool add) {
    if (strcmp(point, "LTorsor") == 0) weapon->mBasefv = mLeft;
    else if (strcmp(point, "LUpArm") == 0) weapon->mBasefv = mLeftupper;
    else if (strcmp(point, "LLoArm") == 0) weapon->mBasefv = mLeftlower;
    else if (strcmp(point, "RTorsor") == 0) weapon->mBasefv = mRight;
    else if (strcmp(point, "RUpArm") == 0) weapon->mBasefv = mRightupper;
    else if (strcmp(point, "RLoArm") == 0) weapon->mBasefv = mRightlower;
    else if (point[0] == 'L') weapon->mBasefv = mLeftupper;
    else if (point[0] == 'R') weapon->mBasefv = mRightupper;
    else weapon->mBasefv = mBack;

    weapon->mOwner = this;
    if (add) mWeapons.push_back(weapon);
}


void cMech::animateWalk(float spf) {
    const float e = 0.0174545;

    float height = checkRelHeight();

    //if (mJetOn >= 50.0f && mJumpstate == FALLING) mJumpstate = JETTING;
    //else if (height > 0.2) mJumpstate = FALLING;
    //else mJumpstate = GROUNDED;

    bool jetting = mJetOn >= 45.0f;
    bool falling = height > 0.5;
    bool landing = !falling;


    if (mJumpstate == GROUNDED && jetting) {
        mJumpstate = JUMPING;
        mVel[Y] += 1.2;
    }

    if (mJumpstate != GROUNDED && landing) mJumpstate = GROUNDED;

    /*
    if (mGrounded) mGrounded = !falling && !jetting;
    else mGrounded = landing && !jetting;
     */

    if (!true && hasRole(HUMANPLAYER)) {
        printf("cMech::animateWalk(): Height over ground %.2f\n", height);
        //printf("Grounded %i\n", mGrounded);
        printf("JetOn: %.2f\n", mJetOn);
    }

    memcpy(mRotators[CTorsor], mTwr, sizeof (float) * 3);

    float speed = -0.5 * mSpeed / 100.0f;

    //if (mGrounded) {
    if (mJumpstate == GROUNDED) {
        float friction = 0.9;
        // No gravity effect as for ground sticking.
        mAcc[X] = mAcc[Y] = mAcc[Z] = 0.0f;

        // Engine Effect
        mAcc[X] += sin(mRot[Y] * e) * cos(mRot[X] * e) * speed;
        mAcc[Z] += cos(mRot[Y] * e) * cos(mRot[X] * e) * speed;

        // Accumulate acceleration and friction force to velocity.

        loopi(3) {
            mVel[i] = friction * mVel[i] + mAcc[i];
        }

        // Correct velocity when colliding
        float ahead = checkAhead(spf);

        // Update Position according to velocity.

        loopi(3) {
            mPos[i] += mVel[i] * spf;
        }

        // Smoothly stick to ground
        mPos[X] -= 0.0f;
        mPos[Y] -= height * 0.4;
        mPos[Z] -= 0.0f;

        // Wobble while moving
        //mPos[X] += 0.05*sp*cos(mRot[Y]*e)*cos(180*mSeconds*e);
        //mPos[Z] += 0.05*sp*sin(mRot[Y]*e)*cos(180*mSeconds*e);

        // Animate Legs
        float t = mSeconds * 180;
        float a = -speed * 1.5;

        float o = 0.017454 * -60;

        float l1 = 30 * (+sin(t * e)*1) * a;
        float l2 = 21 * (+sin(t * e * xsgn(a) + o) + 1) * fabs(a);

        float r1 = 30 * (-sin(t * e)*1) * a;
        float r2 = 21 * (-sin(t * e * xsgn(a) + o) + 1) * fabs(a);

        mRotators[LUpLeg][X] = -l1;
        mRotators[LLoLeg][X] = -l2;
        mRotators[RUpLeg][X] = -r1;
        mRotators[RLoLeg][X] = -r2;

    } else {
        float friction = 0.9;

        // Gravity effect
        float mass = 0.4f;
        mAcc[X] = -cWorld::instance->mGravity[X] * mass;
        mAcc[Y] = -cWorld::instance->mGravity[Y] * mass;
        mAcc[Z] = -cWorld::instance->mGravity[Z] * mass;

        // Jump Jet effect
        mAcc[Y] += mJetOn / 100.0f * 4.2;

        // Engine Effect
        mAcc[X] += sin(mRot[Y] * e) * cos(mRot[X] * e) * speed;
        mAcc[Z] += cos(mRot[Y] * e) * cos(mRot[X] * e) * speed;

        // Accumulate acceleration and friction force to velocity.

        loopi(3) {
            mVel[i] = friction * mVel[i] + mAcc[i];
        }

        // Correct velocity when colliding
        float ahead = checkAhead(spf);

        // Update Position according to velocity.

        loopi(3) {
            mPos[i] += mVel[i] * spf;
        }

        // Hanging Legs
        const float a = -50;
        const float b = -130;
        const float s = 30;
        const float f = b / a;
        if (mRotators[LUpLeg][X] > -a) mRotators[LUpLeg][X] -= s * spf;
        if (mRotators[LUpLeg][X] < -a + 1) mRotators[LUpLeg][X] += s * spf;

        if (mRotators[LLoLeg][X] > +b) mRotators[LLoLeg][X] -= f * s * spf;
        if (mRotators[LLoLeg][X] < +b + 1) mRotators[LLoLeg][X] += f * s * spf;

        if (mRotators[RUpLeg][X] > -a) mRotators[RUpLeg][X] -= s * spf;
        if (mRotators[RUpLeg][X] < -a + 1) mRotators[RUpLeg][X] += s * spf;

        if (mRotators[RLoLeg][X] > +b) mRotators[RLoLeg][X] -= f * s * spf;
        if (mRotators[RLoLeg][X] < +b + 1) mRotators[RLoLeg][X] += f * s * spf;
    }
}

void cMech::animate(float spf) {
    // Update time since spawn
    mSeconds += spf;

    if (mHp[cObject::BODY] <= 75) this->addRole(WOUNDED);
    if (mHp[cObject::BODY] <= 50) this->addRole(SERIOUS);
    if (mHp[cObject::BODY] <= 25) this->addRole(CRITICAL);
    //if (mHp[cObject::BODY] <= 0) this->addRole(DEAD);

    if (hasRole(DEAD)) {
        // Stop movement when dead
        mSpeed = 0;
        mJetOn = 0;
        // If already exploded then frag.
        // Animate Own Explosion (if fired)
        mExplosion.animate(spf);
        if (mExplosion.ready()) {
            this->ChassisLR(0);
            this->TowerLR(0);
            this->TowerUD(0);
            //cWorld::instance->fragObject(this);
        }
        if (mPad) mPad->reset();
    }

    // Process Pad-input when not dead.
    if (!hasRole(DEAD)) {
        assert(mPad != NULL);
        //mPad->print();

        // Any new Messages?
        if (cWorld::instance->mMessages.size() != 0) {
            tSerial last = mLastMessage;

            looplist(i, cWorld::rMessage*, cWorld::instance->mMessages) {
                cWorld::rMessage* m = *i;
                if (m->mSerial > mLastMessage) {
                    last = m->mSerial;
                    Console_printf(mConsole, m->mText.c_str());
                    // If the message contains commands for this
                    // entity then interprete it.
                }
            }
            mLastMessage = last;
        }

        // Let AI-Controller set Pad-input.
        if (mController && mController->mEnabled) {
            if (mPad) mPad->reset();
            mController->process();
        }

        this->TowerLR(50 * spf * -mPad->getAxis(cPad::MECH_TURRET_LR_AXIS));
        this->TowerUD(50 * spf * mPad->getAxis(cPad::MECH_TURRET_UD_AXIS));

        if (!this->hasRole(cObject::IMMOBILE)) {
            // Steer left right
            this->ChassisLR(50 * spf * -mPad->getAxis(cPad::MECH_CHASSIS_LR_AXIS));
            // Speed
            float tdelta = 50 * spf * -mPad->getAxis(cPad::MECH_THROTTLE_AXIS);
            //printf("tdelta: %.2f   speed: %.2f\n", tdelta, mSpeed);
            if (mSpeed > 0 && mSpeed + tdelta < 0) {
                mSpeed = mSpeed * 0.0;
            } else if (mSpeed < 0 && mSpeed + tdelta > 0) {
                mSpeed = mSpeed * 0.0;
            } else {
                mSpeed += 10 * tdelta;
                if (mSpeed > +220) mSpeed = +220;
                else
                    if (mSpeed < -100) mSpeed = -100;
            }
        }

        if (mPad->getButton(cPad::MECH_FIRE_BUTTON)) {
            if (true) {
                mCurrentWeapon %= mWeapons.size();
                fireWeapon(mCurrentWeapon);
                mCurrentWeapon++;
                mCurrentWeapon %= mWeapons.size();
            } else {
                fireAllWeapons();
            }
        }

        if (mPad->getButton(cPad::MECH_JET_BUTTON)) {
            mJetOn += (100 - mJetOn) * 0.1;
        } else {
            mJetOn -= (mJetOn * 0.05);
        }

        if (mPad->getButton(cPad::MECH_CAMERA_BUTTON)) {
            // Only if Camera State is not transitional
            // then switch to next perspective on button press.
            if (mCameraState > 0) {
                mCameraState = (1 + (mCameraState % MAX_CAMERAMODES));
                mCameraState *= -1; // Set transitional.
            }
        } else {
            // Set Camera State as fixed.
            mCameraState = abs(mCameraState);
        }
    }

    // Show fitting animation pose.
    animateWalk(spf);

    // Animate Weapons

    loopi(mWeapons.size()) {
        mWeapons[i]->animate(spf);
    }


    alSourcefv(mSoundSource, AL_POSITION, mPos.data());
    //alSourcefv(mSource, AL_VELOCITY, mVel);
}

/* ----------------------------------------------------------*/
// Model-Rendering-Code (should be evacuated sometime)

void drawLeft(bool transformOnly, float* upper, float* lower) {
    if (upper) {
        glPushMatrix();
        {
            glTranslatef(-0.25, 0, 0);
            glGetFloatv(GL_MODELVIEW_MATRIX, upper);
        }
        glPopMatrix();
    }
    if (lower) {
        glPushMatrix();
        {
            glTranslatef(-0.38, -0.6, -0.0);
            glGetFloatv(GL_MODELVIEW_MATRIX, lower);
        }
        glPopMatrix();
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    glPushMatrix();
    { // Shoulder Model
        glTranslatef(-0.2, 0.0, 0.0);
        glScalef(0.2, 0.1, 0.1);
        glColor3fv(b);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glTranslatef(-0.35, 0, 0);
    //glAxis(0.1);

    glPushMatrix();
    { // Upper Arm Model
        glTranslatef(-0.05, -0.3, -0.15);
        glScalef(0.1, 0.3, 0.3);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glTranslatef(+0.0, -0.4, 0);
    //glAxis(0.1);
    glRotatef(90, 1, 0, 0);
    glTranslatef(+0.0, -0.52, 0);
    //glAxis(0.1);
}

void drawRight(bool transformOnly, float* upper, float* lower) {
    if (upper) {
        glPushMatrix();
        {
            glTranslatef(+0.25, 0, 0);
            glGetFloatv(GL_MODELVIEW_MATRIX, upper);
        }
        glPopMatrix();
    }
    if (lower) {
        glPushMatrix();
        {
            glTranslatef(+0.38, -0.6, -0.0);
            glGetFloatv(GL_MODELVIEW_MATRIX, lower);
        }
        glPopMatrix();
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    glPushMatrix();
    { // Shoulder Model
        glTranslatef(+0.2, 0.0, 0.0);
        glScalef(0.2, 0.1, 0.1);
        glColor3fv(b);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glTranslatef(+0.35, 0, 0);
    //glAxis(0.1);

    glPushMatrix();
    { // Upper Arm Model
        glTranslatef(+0.05, -0.3, -0.15);
        glScalef(0.1, 0.3, 0.3);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();

    glRotatef(0, 1, 0, 0);
    glTranslatef(+0.0, -0.4, 0);
    //glAxis(0.1);
    // Lower Arm

    glRotatef(90, 1, 0, 0);
    glTranslatef(+0.0, -0.52, 0);
    //glAxis(0.1);
    // Hand
}

void drawCenter(bool transformOnly, float xrot, float yrot, float* back, float* left, float* right, float* head) {
    glRotatef(yrot, 0, 1, 0);
    glRotatef(xrot, 1, 0, 0);
    if (head) {
        glPushMatrix();
        {
            glTranslatef(+0.0, +0.97, +0.0);
            glGetFloatv(GL_MODELVIEW_MATRIX, head);
        }
        glPopMatrix();
    }
    if (back) {
        glPushMatrix();
        {
            glTranslatef(+0.0, +0.6, +0.3);
            glGetFloatv(GL_MODELVIEW_MATRIX, back);
        }
        glPopMatrix();
    }
    if (left) {
        glPushMatrix();
        {
            glTranslatef(-0.7, +0.6, +0.2);
            glGetFloatv(GL_MODELVIEW_MATRIX, left);
        }
        glPopMatrix();
    }
    if (right) {
        glPushMatrix();
        {
            glTranslatef(+0.7, +0.6, +0.2);
            glGetFloatv(GL_MODELVIEW_MATRIX, right);
        }
        glPopMatrix();
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    // Center
    glPushMatrix();
    {
        glScalef(0.2, 0.39, 0.4);
        glTranslatef(0, 1, -0.5);
        glColor3fv(b);
        glCenterUnitBlock();
    }
    glPopMatrix();
    // Head
    glPushMatrix();
    {
        glTranslatef(0, 0.94, -0.1);
        glScalef(0.2, 0.16, 0.2);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();
    // Left Center
    glPushMatrix();
    {
        glTranslatef(-0.39, 0.58, 0.0);
        glScalef(0.3, 0.25, 0.35);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();
    // Right Center
    glPushMatrix();
    {
        glTranslatef(+0.39, 0.58, 0.0);
        glScalef(0.3, 0.25, 0.35);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();
}

void drawLower(bool transformOnly, float lhp, float lne, float rhp, float rne, float* center) {
    if (center) {
        glGetFloatv(GL_MODELVIEW_MATRIX, center);
    }
    if (transformOnly) return;

    //glAxis(0.1);

    //float a[] = { 0.3, 0.3, 0.1 };
    //float b[] = { 0.1, 0.1, 0.1 };
    float a[] = {0.9, 0.9, 0.9};
    float b[] = {0.7, 0.7, 0.7};

    glPushMatrix();
    { // Hips Model
        glTranslatef(+0.0, 0.0, 0.0);
        glScalef(0.3, 0.12, 0.12);
        glColor3fv(a);
        glCenterUnitBlock();
    }
    glPopMatrix();

    // Left Leg
    glPushMatrix();
    {
        glTranslatef(-0.32, 0, 0);
        glRotatef(lhp, 1, 0, 0);
        //glAxis(0.1);
        glPushMatrix();
        { // Upper Leg Model
            glTranslatef(-0.0, -0.4, 0.0);
            glScalef(0.18, 0.5, 0.2);
            glColor3fv(b);
            glCenterUnitBlock();
        }
        glPopMatrix();
        glTranslatef(-0, -0.9, 0);
        // Lower Leg
        glPushMatrix();
        {
            glRotatef(lne, 1, 0, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Lower Leg Model
                glTranslatef(-0.0, -0.3, 0.0);
                glScalef(0.1, 0.4, 0.1);
                glColor3fv(a);
                glCenterUnitBlock();
            }
            glPopMatrix();
            glTranslatef(-0, -0.9, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Foot Model
                glTranslatef(0.0, +0.1, -0.09);
                glScalef(0.18, 0.1, 0.26);
                glColor3fv(b);
                glCenterUnitBlock();
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
    glPopMatrix();

    // Right Leg
    glPushMatrix();
    {
        glTranslatef(+0.32, 0, 0);
        glRotatef(rhp, 1, 0, 0);
        //glAxis(0.1);
        glPushMatrix();
        { // Upper Leg Model
            glTranslatef(+0.0, -0.4, 0.0);
            glScalef(0.18, 0.5, 0.2);
            glColor3fv(b);
            glCenterUnitBlock();
        }
        glPopMatrix();
        glTranslatef(+0, -0.9, 0);
        // Lower Leg
        glPushMatrix();
        {
            glRotatef(rne, 1, 0, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Lower Leg Model
                glTranslatef(+0.0, -0.3, 0.0);
                glScalef(0.1, 0.4, 0.1);
                glColor3fv(a);
                glCenterUnitBlock();
            }
            glPopMatrix();
            glTranslatef(+0, -0.9, 0);
            //glAxis(0.1);
            glPushMatrix();
            { // Foot Model
                glTranslatef(0.0, +0.1, -0.09);
                glScalef(0.18, 0.1, 0.26);
                glColor3fv(b);
                glCenterUnitBlock();
            }
            glPopMatrix();
        }
        glPopMatrix();
    }
    glPopMatrix();
}

/* ----------------------------------------------------------*/


void cMech::transform() {
    if (true) {
        glPushMatrix();
        {
            glTranslatef(this->mPos[0], this->mPos[1] + 1.8, this->mPos[2]);
            glRotatef(this->mRot[1], 0, 1, 0);
            if (hasRole(DEAD)) {
                glTranslatef(0, -1.6, 0);
                glRotatef(-40, 4, 1, 3);
            }

            bool transformOnly = true;

            float twx = this->mTwr[0];
            float twy = this->mTwr[1];
            float lhp = this->mRotators[LUpLeg][X];
            float lne = this->mRotators[LLoLeg][X];
            float rhp = this->mRotators[RUpLeg][X];
            float rne = this->mRotators[RLoLeg][X];

            //glAxis(0.15);
            glPushMatrix();
            {
                drawLower(transformOnly, lhp, lne, rhp, rne, mCenter);
                glLoadMatrixf(mCenter);
                drawCenter(transformOnly, twx, twy, mBack, mLeft, mRight, mHead);
                glLoadMatrixf(mLeft);
                drawLeft(transformOnly, mLeftupper, mLeftlower);
                glLoadMatrixf(mRight);
                drawRight(transformOnly, mRightupper, mRightlower);
                glLoadMatrixf(mBack);
                /*
                 * glScalef(0.4,0.4,0.2);
                 * glColor3f(0.4,0,0);
                 * glCenterUnitCylinder(6);
                 */
            }
            glPopMatrix();

        }
        glPopMatrix();
    }

    loopi(mWeapons.size()) {
        mWeapons[i]->transform();
    }
}

void cMech::drawSolid() {
    // Drawing of model should start at the hips as base (0,0,0).
    if (true) {
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        {
            if (ENABLE_TEXTURE_3D) glEnable(GL_TEXTURE_3D);
            glBindTexture(GL_TEXTURE_3D, sTextures[0]);
            glMatrixMode(GL_TEXTURE);
            //glRotatef(0.2,2,3,5);
            //glTranslatef(-0.001,0,0);
            glMatrixMode(GL_MODELVIEW);

            glPushMatrix();
            {
                glTranslatef(this->mPos[0], this->mPos[1] + 1.8, this->mPos[2]);
                glRotatef(this->mRot[1], 0, 1, 0);
                if (hasRole(DEAD)) {
                    glTranslatef(0, -1.6, 0);
                    glRotatef(-40, 4, 1, 3);
                }

                // Local-To-World Matrices (Bases) of Mountpoints.
                float head[16];
                float center[16];
                float back[16];
                float left[16];
                float right[16];
                float rightupper[16];
                float rightlower[16];
                float leftupper[16];
                float leftlower[16];

                bool transformOnly = false;

                float twx = this->mTwr[0];
                float twy = this->mTwr[1];
                float lhp = this->mRotators[LUpLeg][X];
                float lne = this->mRotators[LLoLeg][X];
                float rhp = this->mRotators[RUpLeg][X];
                float rne = this->mRotators[RLoLeg][X];

                //glAxis(0.15);
                glPushMatrix();
                {
                    drawLower(transformOnly, lhp, lne, rhp, rne, center);
                    glLoadMatrixf(center);
                    drawCenter(transformOnly, twx, twy, back, left, right, head);
                    glLoadMatrixf(left);
                    drawLeft(transformOnly, leftupper, leftlower);
                    glLoadMatrixf(right);
                    drawRight(transformOnly, rightupper, rightlower);
                    glLoadMatrixf(back);
                    /*
                glScalef(0.4,0.4,0.2);
                glColor3f(0.4,0,0);
                glCenterUnitCylinder(6);
                     */
                }
                glPopMatrix();

            }
            glPopMatrix();
        }
        glPopAttrib();
    }

    loopi(mWeapons.size()) {
        mWeapons[i]->drawSolid();
    }

    mExplosion.drawSolid();
}
#include <iostream>
void cMech::drawEffect() {
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    if (!mRoles[HUMANPLAYER]) {
        glDisable(GL_CULL_FACE);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glDepthMask(GL_FALSE);
        //glBlendFunc(GL_SRC_ALPHA, GL_ONE);
        if (mRoles[RED]) glColor4f(1,0,0,0.4);
        else if (mRoles[GREEN]) glColor4f(0,1,0,0.4);
        else if (mRoles[BLUE]) glColor4f(0,0,1,0.4);
        else glColor4f(1,1,0,0.4);
        glPushMatrix();
        {
            glMultMatrixf(this->mCenter);
            glTranslatef(0, -1.7, 0);
            glRotatef(90, 1,0,0);
            glDisk(7, 1.3f);
        }
        glPopMatrix();
        glPushMatrix();
        {
            glMultMatrixf(this->mHead);
            glTranslatef(0, +0.9, 0);
            float s = 0.65;
            glScalef(s,s,s);
            float n[16];
            glGetTransposeInverseRotationMatrix(n);
            glMultMatrixf(n);
            int l = mName.length();
            glTranslatef(-l*0.5f,0,0);
            glprintf(mName.c_str());
        }
        glPopMatrix();
    }
    glPopAttrib();

    // Draw jumpjet exaust if jet is somewhat on.
    if (this->mJetOn > 30.0f) {
        glPushAttrib(GL_ALL_ATTRIB_BITS);
        {
            glDisable(GL_CULL_FACE);
            glDisable(GL_LIGHTING);
            glDisable(GL_TEXTURE_2D);
            glDepthMask(GL_FALSE);
            glBlendFunc(GL_SRC_ALPHA, GL_ONE);

            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mCenter);
                glTranslatef(0.3, -0.18, +0.28);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mCenter);
                glTranslatef(-0.3, -0.18, +0.28);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mBack);
                glTranslatef(-0.6, -0.2, +0.1);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
            glPushMatrix();
            {
                float f = mJetOn * 0.003;
                glMultMatrixf(this->mBack);
                glTranslatef(+0.6, -0.2, +0.1);
                float n[16];
                glGetTransposeInverseRotationMatrix(n);
                glMultMatrixf(n);
                glColor4f(1, 1, 0.3, 0.6);
                glDisk(7, f + 0.0003 * (rand() % 100));
                glColor4f(1, 0.5, 0.3, 0.7);
                glDisk(7, f * 0.6 + 0.001 * (rand() % 100));
                glColor4f(1, 1, 1, 0.8);
                glDisk(7, f * 0.3 + 0.001 * (rand() % 100));
            }
            glPopMatrix();
        }
        glPopAttrib();
    }
    // Draw weapon's effects.

    loopi(mWeapons.size()) {
        mWeapons[i]->drawEffect();
    }
    // Draw effect of own explosion (yes, it is only when exploding).
    mExplosion.drawEffect();
}

void cMech::drawHUD() {
    glPushAttrib(GL_ALL_ATTRIB_BITS /* more secure */);
    {
        glLineWidth(2);
        glDisable(GL_LIGHTING);
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_FOG);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_CULL_FACE);

        glPushUnitOrthoProjection();
        {
            glPushMatrix();
            {
                glLoadIdentity();

                // Reticle in screen center.
                if (true && abs(this->mCameraState) == 1) {
                    glPushMatrix();
                    {
                        glTranslatef(0.5, 0.5, 0);
                        glScalef(0.03, 0.04, 1);
                        glPointSize(2);
                        glBegin(GL_POINTS);
                        {
                            glColor4f(1, 0, 0, 0.2);
                            glVertex3f(0, 0, 0);
                        }
                        glEnd();
                        glBegin(GL_LINES);
                        {
                            const float a = 0.7;
                            // left
                            glVertex3f(-1, 0, 0);
                            glVertex3f(-a, 0, 0);
                            // low
                            glVertex3f(0, -1, 0);
                            glVertex3f(0, -a, 0);
                            // right
                            glVertex3f(+1, 0, 0);
                            glVertex3f(+a, 0, 0);
                            // diagonal
                            glVertex3f(-a, 0, 0);
                            glVertex3f(0, -a, 0);
                            glVertex3f(0, -a, 0);
                            glVertex3f(+a, 0, 0);
                        }
                        glEnd();
                    }
                    glPopMatrix();
                }

                // Compass-Lines at top of HUD
                if (true) {
                    glPushMatrix();
                    {
                        glTranslatef(0.5, 0.97, 0);
                        glScalef(1.0, 0.02, 1.0);
                        glColor4f(0, 1, 0, 0.2);

                        float r = this->mRot[Y];
                        if (r > 180) r -= 360;
                        const float d = 1.0f / 360.0f;

                        glLineWidth(3);
                        glBegin(GL_LINES);
                        {
                            glColor4f(0, 0.3, 1, 0.2);
                            float s = 1.2;
                            for (int i = -180; i < 180; i += 10) {
                                float t = fmod(r, 10);
                                glVertex3f((i + t) * d, -s, 0);
                                glVertex3f((i + t) * d, +s, 0);
                            }
                        }
                        glEnd();

                        glLineWidth(1);
                        glBegin(GL_LINES);
                        {
                            glColor4f(1, 1, 1, 0.6);
                            float s = 1.0;
                            for (int i = -180; i < 180; i += 10) {
                                float t = fmod(r, 10);
                                glVertex3f((i + t) * d, -s, 0);
                                glVertex3f((i + t) * d, +s, 0);
                            }
                        }
                        glEnd();

                        glColor4f(1, 0, 0, 0.2);
                        glLineWidth(3);

                        float w = 0.01f;
                        float h = 1.0f;

                        // N for North
                        if (0) {
                            glPushMatrix();
                            {
                                glTranslatef(r*d, 0, 0);
                                glBegin(GL_LINES);
                                {
                                    glVertex3f(-w, +h, 0);
                                    glVertex3f(-w, -h, 0);
                                    glVertex3f(-w, +h, 0);
                                    glVertex3f(+w, -h, 0);
                                    glVertex3f(+w, +h, 0);
                                    glVertex3f(+w, -h, 0);
                                }
                                glEnd();
                            }
                            glPopMatrix();
                        } // north

                        // W for West
                        if (0) {
                            glPushMatrix();
                            {
                                glTranslatef((r - 90) * d, 0, 0);
                                glBegin(GL_LINES);
                                {
                                    glVertex3f(-w, +h, 0);
                                    glVertex3f(-w, -h, 0);
                                    glVertex3f(-w, -h, 0);
                                    glVertex3f(+0, +h * 0.5, 0);
                                    glVertex3f(+0, +h * 0.5, 0);
                                    glVertex3f(+w, -h, 0);
                                    glVertex3f(+w, -h, 0);
                                    glVertex3f(+w, +h, 0);
                                }
                                glEnd();
                            }
                            glPopMatrix();
                        } // west

                        // E for East
                        if (0) {
                            glPushMatrix();
                            {
                                glTranslatef((r + 90) * d, 0, 0);
                                glBegin(GL_LINES);
                                {
                                    glVertex3f(+w, +h, 0);
                                    glVertex3f(-w, +h, 0);
                                    glVertex3f(+w, +0, 0);
                                    glVertex3f(-w, +0, 0);
                                    glVertex3f(+w, -h, 0);
                                    glVertex3f(-w, -h, 0);
                                    glVertex3f(-w, +h, 0);
                                    glVertex3f(-w, -h, 0);
                                }
                                glEnd();
                            }
                            glPopMatrix();
                        } // east

                        // S for South
                        if (0) {
                            glPushMatrix();
                            {
                                glTranslatef((r - 180) * d, 0, 0);
                                glBegin(GL_LINES);
                                {
                                    glVertex3f(+w, +h, 0);
                                    glVertex3f(-w, +h * 0.4, 0);
                                    glVertex3f(-w, +h * 0.4, 0);
                                    glVertex3f(+w, -h * 0.4, 0);
                                    glVertex3f(+w, -h * 0.4, 0);
                                    glVertex3f(-w, -h, 0);
                                }
                                glEnd();
                                glTranslatef((+360) * d, 0, 0);
                                glBegin(GL_LINES);
                                {
                                    glVertex3f(+w, +h, 0);
                                    glVertex3f(-w, +h * 0.4, 0);
                                    glVertex3f(-w, +h * 0.4, 0);
                                    glVertex3f(+w, -h * 0.4, 0);
                                    glVertex3f(+w, -h * 0.4, 0);
                                    glVertex3f(-w, -h, 0);
                                }
                                glEnd();
                            }
                            glPopMatrix();
                        } // south

                    }
                    glPopMatrix();
                } // Compass

                // Ornamental HUD-Border-Frame
                if (true) {
                    glLineWidth(5);
                    glBegin(GL_LINE_STRIP);
                    {
                        glColor4f(1, 1, 0, 0.2);
                        glVertex3f(0.2, 0.1, 0);
                        glVertex3f(0.8, 0.1, 0);
                        glVertex3f(0.9, 0.2, 0);
                        glVertex3f(0.9, 0.8, 0);
                        glVertex3f(0.8, 0.9, 0);
                        glVertex3f(0.2, 0.9, 0);
                        glVertex3f(0.1, 0.8, 0);
                        glVertex3f(0.1, 0.2, 0);
                        glVertex3f(0.2, 0.1, 0);
                    }
                    glEnd();
                    glLineWidth(1);
                    glLineStipple(1, 0xFF55);
                    glEnable(GL_LINE_STIPPLE);
                    glBegin(GL_LINE_STRIP);
                    {
                        glColor4f(1, 1, 1, 0.6);
                        glVertex3f(0.2, 0.1, 0);
                        glVertex3f(0.8, 0.1, 0);
                        glVertex3f(0.9, 0.2, 0);
                        glVertex3f(0.9, 0.8, 0);
                        glVertex3f(0.8, 0.9, 0);
                        glVertex3f(0.2, 0.9, 0);
                        glVertex3f(0.1, 0.8, 0);
                        glVertex3f(0.1, 0.2, 0);
                        glVertex3f(0.2, 0.1, 0);
                    }
                    glEnd();
                    glDisable(GL_LINE_STIPPLE);
                } // Border

                float f = 250;

                // Tower angle indicator
                if (true) {
                    // Tower left/right indication bar.
                    glBegin(GL_LINES);
                    {
                        glColor4f(0, 0.2, 0.7, 0.4);
                        glVertex3f(0.5, 0.1, 0);
                        glVertex3f(0.5 + this->mTwr[1] / f, 0.1, 0);

                        glVertex3f(0.5 + this->mTwr[1] / f, 0.08, 0);
                        glVertex3f(0.5 + this->mTwr[1] / f, 0.12, 0);

                        glVertex3f(0.5, 0.9, 0);
                        glVertex3f(0.5 + this->mTwr[1] / f, 0.9, 0);

                        glVertex3f(0.5 + this->mTwr[1] / f, 0.88, 0);
                        glVertex3f(0.5 + this->mTwr[1] / f, 0.92, 0);
                    }
                    glEnd();

                    // Tower up/down indication bar.
                    glBegin(GL_LINES);
                    {
                        glColor4f(0, 0.2, 0.7, 0.4);
                        glVertex3f(0.1, 0.5, 0);
                        glVertex3f(0.1, 0.5 + this->mTwr[0] / f, 0);

                        glVertex3f(0.08, 0.5 + this->mTwr[0] / f, 0);
                        glVertex3f(0.12, 0.5 + this->mTwr[0] / f, 0);

                        glVertex3f(0.9, 0.5, 0);
                        glVertex3f(0.9, 0.5 + this->mTwr[0] / f, 0);

                        glVertex3f(0.88, 0.5 + this->mTwr[0] / f, 0);
                        glVertex3f(0.92, 0.5 + this->mTwr[0] / f, 0);
                    }
                    glEnd();

                } // Tower angle indicator


                // Weapon Status "Buttons"
                if (true) {
                    glPushMatrix();
                    {
                        //glLoadIdentity();
                        glScalef(0.2, 0.05, 1.0);

                        loopi(mWeapons.size()) {
                            glLineWidth(5);
                            if (mWeapons[i]->ready()) glColor4f(.4, 1, .4, 0.2);
                            else glColor4f(.8, 0, 0, 0.2);
                            glLineSquare(0.1);
                            glLineWidth(1);
                            glColor4f(1, 1, 1, 0.9);
                            glLineSquare(0.1);
                            glPushMatrix();
                            {
                                mWeapons[i]->drawHUD();
                            }
                            glPopMatrix();
                            glTranslatef(0, 1, 0);
                        }
                    }
                    glPopMatrix();
                } // Weapons status


                // Damage indicator boxes
                if (true) {
                    glPushMatrix();
                    {
                        for (int i = 1; i <= 5; i += 4) {
                            if (i == 1) glColor4f(1, 1, 1, 0.9);
                            glLineWidth(i);
                            glLoadIdentity();
                            glTranslatef(0.8, 0.1, 0);
                            glScalef(0.06, 0.08, 1.0);
                            // Left Arm
                            if (i != 1) glColor4f(1 - this->mHp[cObject::LEFT]*0.01, this->mHp[cObject::LEFT]*0.01, 0.4, 0.2);
                            glLineSquare(0.1);
                            // Torsor&Head
                            glTranslatef(1, 0, 0);
                            glScalef(1, 1.5, 1);
                            if (i != 1) glColor4f(1 - this->mHp[cObject::BODY]*0.01, this->mHp[cObject::BODY]*0.01, 0.4, 0.2);
                            glLineSquare(0.1);
                            // Right Arm&Shoulder
                            glTranslatef(1, 0, 0);
                            glScalef(1, 1.0f / 1.5f, 1);
                            if (i != 1) glColor4f(1 - this->mHp[cObject::RIGHT]*0.01, this->mHp[cObject::RIGHT]*0.01, 0.4, 0.2);
                            glLineSquare(0.1);
                            // Legs
                            glTranslatef(-1.6, -1, 0);
                            glScalef(2.2, 1, 1);
                            if (i != 1) glColor4f(1 - this->mHp[cObject::LEGS]*0.01, this->mHp[cObject::LEGS]*0.01, 0.4, 0.2);
                            glLineSquare(0.1);
                        }
                    }
                    glPopMatrix();
                } // Damage indicator


                // Radar
                if (true) {
                    glPushMatrix();
                    {
                        glScalef(0.1, 0.1, 1);
                        glTranslatef(1.4, 8.6, 0);
                        glDisk(8, 1.1);
                        glRotatef(-this->mRot[1], 0, 0, 1);
                        std::list<cObject*>* near = cWorld::instance->filterByRange(this, this->mPos.data(), 0, 150, -1, NULL);
                        glPointSize(4);
                        glBegin(GL_POINTS); {
                            glColor4f(1,1,1,1);
                            glVertex3f(0,0,0);
                        } glEnd();
                        looplist(i, cObject*, *near) {
                            glBegin(GL_POINTS); {
                                cObject* o = *i;
                                if (o->mRoles[RED]) glColor4f(1,0,0,1);
                                else if (o->mRoles[GREEN]) glColor4f(0,1,0,1);
                                else if (o->mRoles[BLUE]) glColor4f(0,0,1,1);
                                else if (o->mRoles[YELLOW]) glColor4f(0,1,0,1);
                                else glColor4f(0.5,0.5,0.5,1);
                                float dx = o->mPos[0]-this->mPos[0];
                                float dz = o->mPos[2]-this->mPos[2];
                                float r = sqrt( dx*dx + dz*dz );
                                float f = 0.2 * log(r);
                                glVertex3f(f*dx / r, -f*dz / r, 0);
                            } glEnd();
                        }
                        delete near;
                    }
                    glPopMatrix();
                } // Radar

                // Message Box and Text
                if (true) {
                    cWorld* w = cWorld::instance;
                    glPushMatrix();
                    if (0) {
                        glColor4f(0.09, 0.09, 0.99, 0.1);
                        glTranslatef(0.5, 0, 0);
                        glLineSquare(0.01);
                    }
                    glPopMatrix();

                    glPushMatrix();
                    {
                        glColor4f(0.99, 0.99, 0.19, 1);
                        glTranslatef(1.0f / 80.0f * 15, 1.0f / 40.0f * (9 + 1), 0);
                        glScalef(1.0f / 80.0f, 1.0f / 40.0f, 1.0f);
                        Console_draw(mConsole);
                    }
                    glPopMatrix();

                    glPushMatrix();
                    {
                        glColor4f(0.99, 0.99, 0.19, 1);
                        glTranslatef(0, 1, 0);
                        glScalef(1.0f / 80.0f, 1.0f / 40.0f, 1.0f);
                        glColor4f(0.09, 0.99, 0.09, 1);
                        //glRotatef(1, 0,0,1);
                        glTranslatef(22, -3, 0);
                        glprintf("MKS: %2i (%2i)", int(64 - this->mRot[1] * 64 / 360.0f) % 64, int(64 - (this->mRot[1] + this->mTwr[1]) * 64 / 360.0f) % 64);
                        glTranslatef(14, 0, 0);
                        glprintf("Mercator: %06i %06i", int(this->mPos[0]), int(this->mPos[2]));
                        glTranslatef(-10, -1, 0);
                    }
                    glPopMatrix();

                    glPushMatrix();
                    {
                        glRotatef(90, 0, 0, 1);
                        glScalef(1.0f / 80.0f, 1.0f / 40.0f, 1.0f);
                        glTranslatef(28, -3, 0);
                        glprintf("Date: %i_%i Time: %i:%02i:%02i", w->mYear, w->mDay, w->mHour, w->mMinute, w->mSecond);
                    }
                    glPopMatrix();

                    glPushMatrix();
                    {
                        glColor4f(0.79, 0.79, 0.99, 1);
                        glRotatef(-90, 0, 0, 1);
                        glScalef(1.0f / 80.0f, 1.0f / 40.0f, 1.0f);
                        glTranslatef(-80 + 16, 40 - 3, 0);
                        glprintf("hackcraft.de  - powering e to the mc square");
                    }
                    glPopMatrix();
                } // Message Box and Text

            }
            glPopMatrix();
        }
        glPopProjection();
    }
    glPopAttrib();
}

cObject::rImpact* cMech::damageByBeam(float* a, float* b, float radius, float damage, bool all) {
    cObject::rImpact *result = NULL;
    float* p = mPos.data();
    float* r = mRot.data();
    glPushMatrix();
    glLoadIdentity();
    glRotatef(-r[1], 0, 1, 0);
    glTranslatef(-p[0], -p[1], -p[2]);
    float M[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, M);
    float a_local[3];
    applyMatrix3(a_local, M, a);
    float b_local[3];
    applyMatrix3(b_local, M, b);
    float mins[3] = {-1.0, 0, -0.6};
    float maxs[3] = {+1.0, +3, +0.6};
    float impact[3];
    float normal[3];
    float delta = collideLineWithAABB(a_local, b_local, mins, maxs, impact, normal);
    if (delta >= 0.0f) {
        //printf("cMech::damageByBeam(): hit\n");
        //loopi(MAX_PARTS) { this->mHp[i] -= damage; }
        this->mHp[cObject::BODY] -= damage;
        if (this->mHp[cObject::BODY] <= 0 && !this->hasRole(cObject::DEAD)) {
            this->remRole(DAMAGEABLE);
            this->remRole(COLLIDEABLE);
            this->addRole(DEAD);
            printf("cMech::damageByBeam(): DEAD\n");
            mExplosion.fire(NULL);
        }

        glLoadIdentity();
        glTranslatef(p[0], p[1], p[2]);
        glRotatef(r[1], 0, 1, 0);
        glGetFloatv(GL_MODELVIEW_MATRIX, M);

        result = new cObject::rImpact;
        applyMatrix3(result->pos, M, impact);
        M[12] = M[13] = M[14] = 0.0f; // W/out translation for normal
        applyMatrix3(result->nrm, M, normal);
    }
    glPopMatrix();
    return result;
}

tSerial cMech::enemyNearby() {
    tSerial result = 0;
    // Filter objects by distance
    std::list<cObject*>* scan = cWorld::instance->filterByRange(this, this->mPos.data(), 0.0f, 50.0f, -1, NULL);
    if (!scan->empty()) {
        // Find all objects belonging to any enemy party/role.
        std::list<cObject*>* roles = cWorld::instance->filterByRole(this, this->mEnemies.to_ulong(), false, -1, scan);
        if (!roles->empty()) {
            // Only those objects which are damageable.
            std::list<cObject*>* damages = cWorld::instance->filterByRole(this, (1UL << cObject::DAMAGEABLE), false, 1, roles);
            //cWorld::instance->printNames(damages);
            if (!damages->empty()) result = damages->front()->mSerial;
            delete damages;
        }
        delete roles;
    }
    delete scan;
    return result;
}

float cMech::inDestinationRange() {
    float a = 0;
    float b = 6;
    float d = lineDistance(mPos, mDestination);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

float cMech::inMeeleRange() {
    cObject* target = cWorld::instance->mIndex[mTarget];
    if (target == NULL) return 0.0f;
    float a = 16;
    float b = 24;
    float d = lineDistance(mPos, target->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

float cMech::inWeaponRange() {
    // TODO should depend on ready to fire weapons properties.
    cObject* target = cWorld::instance->mIndex[mTarget];
    if (target == NULL) return 0.0f;
    float a = 36 + 10;
    float b = 44 + 10;
    float d = lineDistance(mPos, target->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

float cMech::inTargetRange() {
    cObject* target = cWorld::instance->mIndex[mTarget];
    if (target == NULL) return 0.0f;
    float a = 56;
    float b = 64;
    float d = lineDistance(mPos, target->mPos);
    if (d < a) return 1.0f;
    if (d > b) return 0.0f;
    return (1.0f - (d - a) / (b - a));
}

// Calculate nearest signed rotation angle towards target position.
// First and second axis should be X (0) and Z (2) for angle around Y (1).
#define rotationTo(own_pos, target_pos, own_angle, first_axis, second_axis) \
({ \
    float r = 0.0f; \
    if (target_pos != NULL) { \
        const float e = 180.0f/3.145479f; \
        float dir[3] = { \
            (own_pos)[0] - (target_pos)[0], \
            (own_pos)[1] - (target_pos)[1], \
            (own_pos)[2] - (target_pos)[2]  \
        }; \
        float tgt_angle = atan2(dir[first_axis], dir[second_axis]) * e; \
        float delta = tgt_angle - (own_angle); \
        r = delta; \
        if (fabs(delta) > 180) r = -(delta / fabs(delta)) * (360 - fabs(delta)); \
        if (!true) printf("own %.2f   tgt %.2f   delta %.2f  r %.2f\n", own_angle, tgt_angle, delta, r); \
        r *= -1.0f; \
    } \
    r; \
})

void cMech::do_moveTowards() {
    //printf("do_moveTowards():\n");
    if (mPad == NULL) return;

    // Determine Target (ie. Move- or Aim-Target if former is not available).
    float* target_pos = NULL;
    if (finitef(mDestination[0])) {
        target_pos = mDestination.data();
    } else if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) {
            target_pos = target->mPos.data();
        } else {
            mTarget = 0;
            return;
        }
    } else return;

    // Determine nearest rotation direction
    float rd = rotationTo(mPos, target_pos, this->mRot[Y], X, Z);
    mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, rd);

    float thr = 20.0f * (1.0f - fabs(rd) / 360.0f);
    //printf("Throttle: %.2f\n", thr);
    mPad->setAxis(cPad::MECH_THROTTLE_AXIS, (char) - thr);
}

void cMech::do_moveNear() {
    //printf("do_moveNear():\n");
    if (mPad == NULL) return;

    // Determine Target (ie. Move- or Aim-Target if former is not available).
    float* target_pos = NULL;
    if (finitef(mDestination[0])) {
        target_pos = mDestination.data();
    } else if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) {
            target_pos = target->mPos.data();
        } else {
            mTarget = 0;
            return;
        }
    } else return;

    // Determine nearest rotation direction
    float rd = rotationTo(mPos, target_pos, this->mRot[Y], X, Z);
    mPad->setAxis(cPad::MECH_CHASSIS_LR_AXIS, rd);

    float d = lineDistance(this->mPos, target_pos);

    float f = (d - 11);
    if (f < -1) f = -1;
    //if (f < 1) f = 0;
    if (f > +5) f = +5;
    //float f = d;
    mPad->setAxis(cPad::MECH_THROTTLE_AXIS, -f * 0.2f);
}

void cMech::do_aimAt() {
    //printf("do_aimAt():\n");
    if (mPad == NULL) return;

    // Get aim-target position.
    float* target_pos = NULL;
    if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) target_pos = target->mPos.data();
        else return;
    } else return;

    // Determine nearest rotation direction
    float rd = rotationTo(mPos, target_pos, this->mRot[Y] + this->mTwr[Y], X, Z);
    mPad->setAxis(cPad::MECH_TURRET_LR_AXIS, rd / 2.0f);
}

void cMech::do_fireAt() {
    //printf("do_fireAt():\n");
    if (mPad == NULL) return;

    // Get aim-target position.
    float* target_pos = NULL;
    if (mTarget != 0) {
        cObject* target = cWorld::instance->mIndex[mTarget];
        if (target != NULL) target_pos = target->mPos.data();
        else return;
    } else return;

    // Determine nearest rotation direction
    float rd = rotationTo(mPos, target_pos, this->mRot[Y] + this->mTwr[Y], X, Z);
    // Fire at random and only if angle small enough.
    if (rand() % 6 == 0 && fabs(rd) < 3) {
        mPad->setButton(cPad::MECH_FIRE_BUTTON, true);
        //printf("fire\n");
    }
}

void cMech::do_idle() {
    if (mPad == NULL) return;
    mSpeed = 0;
    mPad->reset();
}

void cMech::do_aimFor(tSerial target) {
    mTarget = target;
}

void cMech::do_moveFor(float* dest) {
    if (dest != NULL) {
        //mTarget = NULL;
        memcpy(mDestination.data(), dest, sizeof (float) *3);
    } else {
        // Set XYZ to Not-A-Number (NaN) for no location.
        // Note that NaN-ity can only be tested either by
        // isnanf(x), !finite(x) or by x!=x as NaN always != NaN.
        mDestination[0] = nanf("");
        mDestination[1] = nanf("");
        mDestination[2] = nanf("");
        //printf("Destination is %s\n", (finitef(mDestination[0])) ? "finite" : "infinite" );
    }
}

