// cWorld.cpp

#include "cWorld.h"

#include "cBackground.h"
#include "cObject.h"
#include "cMission.h"

#include <GL/gl.h>
#include <iostream>

#define loopObjects(objset)   for (std::list<cObject *>::iterator it = (objset).begin(); it != (objset).end(); it++)
#define loopObjectsNoInc(objset)   for (std::list<cObject *>::iterator it = (objset).begin(); it != (objset).end();)


void drawKDTree(kdtree<cObject*>::KDNode* node, int curlvl, float* min, float* max) {
    int a = curlvl % 3; // divisioning axis 0,1,2?
    bool two = true; // two dimensional kd?
    if (two) a = (curlvl % 2) * 2; // divisioning axis 0,2?

    int b = (a != 0) ? 0 : 1; // other axis
    int c = (a != 1 && b != 1) ? 1 : 2; // remaining axis

    //printf("%i %i %i (%.2f)\n", a, b, c, node->comparevalue);

    //if (node->dataset) std::cout << node->dataset->size() << std::endl;

    if (curlvl < 100) {
        float min_[3];
        float max_[3];
        memcpy(max_, max, sizeof (float) *3);
        max_[a] = node->comparevalue - 0.00001;
        if (node->left) drawKDTree(node->left, curlvl + 1, min, max_);

        memcpy(min_, min, sizeof (float) *3);
        min_[a] = node->comparevalue;
        if (node->right) drawKDTree(node->right, curlvl + 1, min_, max);
    }

    float pnt[4][3];
    for (int i = 0; i < 4; i++) pnt[i][a] = node->comparevalue;

    pnt[0][b] = min[b];
    pnt[0][c] = min[c];
    pnt[1][b] = min[b];
    pnt[1][c] = max[c];
    pnt[2][b] = max[b];
    pnt[2][c] = max[c];
    pnt[3][b] = max[b];
    pnt[3][c] = min[c];

    glColor4f((a == 0) ? 1 : 0.3, (a == 1) ? 1 : 0.3, (a == 2) ? 1 : 0.3, 0.1 + 0.7 / (1.0f + 0.3*curlvl));
    glBegin(GL_QUADS);
    if (a!= 1) {
        for (int i = 0; i < 4; i++) {
            glVertex3fv(pnt[i]);
            //printf("%.2f  %.2f %.2f\n", pnt[i][0], pnt[i][1], pnt[i][2]);
        }
    }
    glEnd();
    //printf("\n");
}


cWorld* cWorld::instance = NULL;

cWorld::cWorld() {
    cWorld::instance = this;

    mBackground = NULL;
    mMission = NULL;

    mGravity.reserve(3);
    mGravity.resize(3);
    mGravity[0] = 0;
    mGravity[1] = 9.8f;
    mGravity[2] = 0;

    mFPS = 30.0f;
    mSPF = 1.0f / mFPS;

    mYear = 2009;
    mDay = 125;
    mHour = 01;
    mMinute = 01;
    mSecond = 01;
    mMSec = 0;
    mFrame = 0;
    mDeltacycle = 1;

    mCluster = NULL;

    //getSystemHour(mTime);
}


// Querries

tSerial cWorld::getSerial() {
    tSerial ddd = mDay;
    tSerial hh = mHour;
    tSerial mm = mMinute;
    tSerial ss = mSecond;
    tSerial ff = mFrame;
    tSerial serid = ((((ddd * 100 + hh) *100 + mm) *100 + ss) * 100 + ff) * 10000 + mDeltacycle;
    return serid;
}


// Messaging

void cWorld::sendMessage(const char* format, ...) {
    FORMATBUFFER
    sendMessageT(buffer);
}

void cWorld::sendMessageT(const char* text) {
    cWorld::rMessage* m = NULL;
    m = new cWorld::rMessage;
    m->mSerial = getSerial();
    m->mText = text;
    cWorld::instance->mMessages.push_back(m);
}


// Spawning And Fragging

void cWorld::spawnObject(cObject *object) {
    tSerial serid = getSerial();

    std::cout << serid << " spawned at deltacycle " << mDeltacycle << std::endl;

    mIndex[serid] = object;
    object->mSerial = serid;
    mDeltacycle++;

    mObjects.push_back(object);
    object->onSpawn();
}

void cWorld::fragObject(cObject *object) {
    if (object->hasRole(cObject::FRAGGED)) return;
    object->addRole(cObject::FRAGGED);
    mObjects.remove(object);
    mCorpses.push_back(object);
    mIndex.erase(object->mSerial);
    object->onFrag();
}

void cWorld::bagFragged() {
    // Finally delete objects which where fragged before.
    while (!mCorpses.empty()) {
        cObject* object = mCorpses.front();
        mCorpses.pop_front();
        delete object;
    };
}


// Simulation Step

void cWorld::advanceTime(int deltamsec) {
    if (mMission) mMission->checkConditions();
    mSPF = float(deltamsec) / 1000.0f;
    mFPS = 1.0 / (mSPF + 0.000001f);

    mFrame++;
    mMSec += deltamsec;
    if (mMSec >= 1000) {
        mFrame = 0;
        mSecond += mMSec / 1000;
        mMSec %= 1000;
        if (mSecond >= 60) {
            mMinute += mSecond / 60;
            mSecond %= 60;
            if (mMinute >= 60) {
                mHour += mMinute / 60;
                mMinute %= 60;
                if (mHour >= 24) {
                    mDay += mHour / 24;
                    mHour %= 24;
                    if (mDay >= 366) {
                        mYear += mDay / 366;
                        mDay %= 366;
                    }
                }
            }
        }
    }

    mDeltacycle = 1;
    tSerial serid = getSerial();
    //std::cout << serid << ": " << mHour << " " << mMinute << " " << mSecond << " " mDeltacycle << std::endl;

    // Update Cluster for acceleration of spatial searches:

    std::list<float*> keys;
    std::list<cObject*> values;

    static float (*p2)[2] = NULL;
    if (p2) delete[] p2;
    p2 = new float[mObjects.size()][2];
    int j = 0;
    looplist(i, cObject*, mObjects) {
        p2[j][0] = (*i)->mPos[0];
        p2[j][1] = (*i)->mPos[2];
        keys.push_back(p2[j]);
        values.push_back(*i);
        j++;
    }
    if (mCluster) delete mCluster;
    mCluster = new kdtree<cObject*>(2, &keys, &values);
    keys.clear();
    values.clear();
}

void cWorld::animateObjects() {
    float spf = mSPF;

    loopObjectsNoInc(mObjects) {
        cObject* object = *it++;
        object->animate(spf);
    };
}

void cWorld::transformObjects() {

    loopObjectsNoInc(mObjects) {
        cObject* object = *it++;
        glPushMatrix();
        object->transform();
        glPopMatrix();
    };
}

void cWorld::drawBack() {
    if (mBackground) mBackground->drawBackground(float(mHour) + float(mMinute) / 60.0f);
}

void cWorld::drawSolid(cObject* camera) {
    float* origin = camera->mPos.data();
    float maxrange = 80 + 25;
    //float min[] = {origin[0] - maxrange, -10000, origin[2] - maxrange};
    //float max[] = {origin[0] + maxrange, +10000, origin[2] + maxrange};
    float min[] = {origin[0] - maxrange, origin[2] - maxrange};
    float max[] = {origin[0] + maxrange, origin[2] + maxrange};
    std::list<cObject*>* objects = mCluster->findUsersetInterval(min, max);
    //std::cout << "vis:" << objects->size() << " vs " << mObjects.size() << std::endl;

    //loopObjectsNoInc(mObjects) {

    loopObjectsNoInc(*objects) {
        cObject* object = *it++;
        glPushMatrix();
        object->drawSolid();
        glPopMatrix();
    };
}


void cWorld::drawEffect(cObject* camera) {
    float* origin = camera->mPos.data();
    //float maxrange = 80 + 25;
    float maxrange = 80 + 25;
    //float min[] = {origin[0] - maxrange, -10000, origin[2] - maxrange};
    //float max[] = {origin[0] + maxrange, +10000, origin[2] + maxrange};
    float min[] = {origin[0] - maxrange, origin[2] - maxrange};
    float max[] = {origin[0] + maxrange, origin[2] + maxrange};
    std::list<cObject*>* objects = mCluster->findUsersetInterval(min, max);
    //    std::cout << "vis:" << objects->size() << " vs " << mObjects.size() << std::endl;
    //std::cout << "clus:" << mCluster->size() << " vs " << mObjects.size() << std::endl;

    //loopObjectsNoInc(mObjects) {

    loopObjectsNoInc(*objects) {
        cObject* object = *it++;
        glPushMatrix(); {
            object->drawEffect();
        } glPopMatrix();
    }

    delete objects;

    glPushAttrib(GL_ALL_ATTRIB_BITS);
    {
        glDisable(GL_CULL_FACE);
        glDisable(GL_TEXTURE_2D);
        glDisable(GL_TEXTURE_3D);
        float d = 600;
        min[0] = -d;
        min[1] = +1;
        min[2] = -d;
        max[0] = +d;
        max[1] = +40;
        max[2] = +d;
        glPushMatrix(); {
            //drawKDTree(mCluster->root, 0, min, max);
        } glPopMatrix();
    }
    glPopAttrib();
    
}


// World Scanning And Filtering

void cWorld::printNames(std::list<cObject*>* objects) {
    char buffer[1024];
    std::string s = "[";
    if (objects == NULL) objects = &mObjects;

    loopObjects(*objects) {
        sprintf(buffer, "%llu", (*it)->mSerial);
        s += " ";
        s += ((*it)->mName);
        s += "#";
        s += buffer;
        s += "(";
        sprintf(buffer, "%lu", (*it)->mRoles.to_ulong());
        s += buffer;
        s += ")";
    }
    s += " ]";
    printf("%s\n", s.c_str());
}

std::list<cObject*>* cWorld::filterByRole(cObject* ex, unsigned long rolemask, bool all, int maxamount, std::list<cObject*>* objects) {
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    if (objects == NULL) objects = &mObjects;

    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object->mRoles[cObject::FRAGGED]) continue;
        // Filter Condition
        unsigned long intersection = object->mRoles.to_ulong() & rolemask;
        if ((intersection != rolemask) && !((intersection != 0) && !all)) continue;
        if (object == ex) continue;
        amount--;
        result->push_back(object);
    }
    return result;
}

std::list<cObject*>* cWorld::filterByRange(cObject* ex, float* origin, float minrange, float maxrange, int maxamount, std::list<cObject*>* objects) {
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    bool del = false;
    if (objects == NULL) {
        //objects = &mObjects;
        del = true;
        //float min[] = {origin[0] - maxrange - 25, -10000, origin[2] - maxrange - 25};
        //float max[] = {origin[0] + maxrange + 25, +10000, origin[2] + maxrange + 25};
        float maxrange_ = maxrange + 1;
        float min[] = {origin[0] - maxrange_, origin[2] - maxrange_};
        float max[] = {origin[0] + maxrange_, origin[2] + maxrange_};
        objects = mCluster->findUsersetInterval(min, max);
        //std::cout << "clustered:" << objects->size() << " vs " << mObjects.size() << std::endl;
    }

    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object == ex) continue;
        if (object->mRoles[cObject::FRAGGED]) continue;
        // Filter Condition
        float d = lineDistance(origin, object->mPos);
        if ((d > maxrange + object->mRadius) || (d < minrange - object->mRadius)) continue;
        amount--;
        result->push_back(object);
    }
    if (del) delete objects;
    return result;
}

std::list<cObject*>* cWorld::filterByName(cObject* ex, char* name, int maxamount, std::list<cObject*>* objects) {
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    if (objects == NULL) objects = &mObjects;

    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object == ex) continue;
        if (object->mRoles[cObject::FRAGGED]) continue;
        // Filter Condition
        if (strcmp(object->mName.c_str(), name) != 0) continue;
        amount--;
        result->push_back(object);
    }
    return result;
}

std::list<cObject*>* cWorld::filterByBeam(cObject* ex, float* pointa, float* pointb, float radius, int maxamount, std::list<cObject*>* objects) {
    float* x1 = pointa;
    float* x2 = pointb;
    std::list<cObject*>* result = new std::list<cObject*>;
    int amount = maxamount;
    if (objects == NULL) objects = &mObjects;

    loopObjects(*objects) {
        if (amount == 0) break;
        cObject *object = *it;
        if (object == ex) continue;
        if (object->mRoles[cObject::FRAGGED]) continue;
        // Filter Condition

        float* x0 = object->mPos.data();
        float u[] = {x0[0] - x1[0], x0[1] - x1[1], x0[2] - x1[2]};
        float v[] = {x2[0] - x1[0], x2[1] - x1[1], x2[2] - x1[2]};
        float a = u[0] * v[0] + u[1] * v[1] + u[2] * v[2];
        float b = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
        float delta = a / b;

        //printf("delta: %.2f\n", delta);

        if (delta < -0.01) continue;
        if (delta > +1.01) continue;

        float p[] = {
            x1[0] + delta * (x2[0] - x1[0]),
            x1[1] + delta * (x2[1] - x1[1]),
            x1[2] + delta * (x2[2] - x1[2])
        };

        float dv[] = {
            p[0] - x0[0],
            p[1] - x0[1],
            p[2] - x0[2]
        };

        float d2 = dv[0] * dv[0] + dv[1] * dv[1] + dv[2] * dv[2];
        float is = radius + object->mRadius;

        if (d2 > is * is) continue;

        amount--;
        result->push_back(object);
    }
    return result;
}


